# -*- coding: utf-8 -*-

import pkg_resources
from slapos.util import bytes2str

with pkg_resources.resource_stream('slapos.proxy', 'schema.sql') as f:
  DB_VERSION = bytes2str(f.readline()).strip().split(':')[1]




import configparser
import json
import logging
import os
import time
import unittest
import xml.etree.ElementTree as ET
import requests
import re
import slapos.client
from slapos.slap import ResourceNotReady
import urllib.parse

from logging.handlers import RotatingFileHandler


class EndToEndTestCase(unittest.TestCase):

  @classmethod
  def setUpClass(cls):
    cls.createLogger()

    configp = configparser.ConfigParser()
    config_path = os.environ['SLAPOS_E2E_TEST_CLIENT_CFG']
    configp.read(config_path)
    args = type("empty_args", (), {})()
    conf = slapos.client.ClientConfig(args, configp)
    local = slapos.client.init(conf, logging.getLogger(__name__))
    cls.slap = local['slap']
    cls.supply = staticmethod(local['supply'])
    cls._request = staticmethod(local['request'])
    cls._requested = {}
    cls._supplied = {}

  @classmethod
  def createLogger(cls):

    LOG_FILE = os.environ['SLAPOS_E2E_TEST_LOG_FILE']
    if not LOG_FILE:
        raise EnvironmentError("SLAPOS_E2E_TEST_LOG_FILE environment variable not set")

    cls.logger = logging.getLogger('logger')
    cls.logger.setLevel(logging.DEBUG)
    handler = RotatingFileHandler(LOG_FILE, maxBytes=100000, backupCount=5)
    cls.logger.addHandler(handler)
    formatter = logging.Formatter("%(asctime)s [%(levelname)s] %(message)s")
    handler.setFormatter(formatter)

  @classmethod
  def tearDownClass(cls):
    for args, kw in cls._requested.values():
      kw['state'] = 'destroyed'
      cls._request(*args, **kw)

    # remove software releases
    for k,v in cls._supplied.items():
      print("Remove %s from %s" %(k, v))
      cls._supply(k, v, 'destroyed')


  @classmethod
  def request(cls, *args, **kw):
    instance_name = args[1]
    cls._requested[instance_name] = (args, kw)
    partition = cls._request(*args, **kw)
    try:
      return cls.unwrapConnectionDict(partition.getConnectionParameterDict())
    except ResourceNotReady:
      # Instance has not yet been allocated, and could not retrieve any info
      return {}

  @classmethod
  def supply(cls, software_release, computer_id, state):
    cls._supplied[software_release] = computer_id
    cls._supply(software_release, computer_id, state)

  @staticmethod
  def unwrapConnectionDict(connection_dict):
    try:
      connection_json = connection_dict['_']
    except KeyError:
      return connection_dict
    return json.loads(connection_json)

  @classmethod
  def getInstanceInfos(cls, instance_name):
    # adapated from cli/info
    infos = cls.slap.registerOpenOrder().getInformation(instance_name)

    class Infos:

      def __init__(self, **kw):
        self.__dict__.update(kw)

    connection_dict = {
      e['connection_key']: e['connection_value']
      for e in infos._connection_dict
    }
    return Infos(
      software_url=infos._software_release_url,
      software_type=infos._source_reference,
      shared=bool(infos._root_slave),
      requested_state=infos._requested_state,
      parameter_dict=infos._parameter_dict,
      connection_dict=cls.unwrapConnectionDict(connection_dict),
      news=infos._news,
    )

  @classmethod
  def getInstanceNews(cls, instance_name):
    try:
      news = cls.slap.registerOpenOrder().getInformation(instance_name)._news
    except Exception:
      return ()
    return news['instance']

  @classmethod
  def getInstanceStatus(cls, instance_name):
    # adapted from cli/info
    status = 0b00
    for e in cls.getInstanceNews(instance_name):
      text = e.get('text', '')
      if text.startswith('#access'):
        status |= 0b01
      elif text.startswith('#error'):
        status |= 0b10
      if status == 0b11:
        break
    return ('none', 'green', 'red', 'orange')[status]

  @classmethod
  def checkTimeoutAndSleep(cls, t0, timeout, msg, interval=60):
    if (time.time() - t0) > 60 * timeout:
      raise TimeoutError(msg)
    time.sleep(interval)

  @classmethod
  def waitUntilGreen(cls, instance_name, timeout=80, t0=None):
    t0 = t0 or time.time()
    while (status := cls.getInstanceStatus(instance_name)) != 'green':
      msg = 'Instance %s status is still %s' % (instance_name, status)
      cls.logger.info(msg)
      cls.checkTimeoutAndSleep(t0, timeout, msg)

  @classmethod
  def waitUntilPublished(cls, instance_name, key, timeout=80, t0=None):
    t0 = t0 or time.time()
    msg = 'Instance %s still does not publish %s' % (instance_name, key)
    while (value := cls.getInstanceInfos(instance_name).connection_dict.get(key)) is None:
        cls.logger.info(msg)
        cls.checkTimeoutAndSleep(t0, timeout, msg)

    return value

  @classmethod
  def waitUntilMonitorURLReady(
      cls, instance_name, code=200, timeout=80, t0=None):
    suffix = '/share/public/feed'
    key = 'monitor-base-url'
    t0 = t0 or time.time()
    url = cls.waitUntilPublished(instance_name, key, timeout, t0) + suffix
    while True:
      resp = requests.get(url, verify=False)
      if resp.status_code == code:
        return resp, url
      msg = 'Instance %s monitor url %s returned unexpected status code %s (expected %s)' % (
        instance_name, url, resp.status_code, code)
      cls.checkTimeoutAndSleep(t0, timeout, msg)
      url = cls.getInstanceInfos(instance_name).connection_dict.get(
        key) + suffix

  @classmethod
  def waitUntilPrivateMonitorURLReady(
      cls, instance_name, code=200, timeout=80, t0=None):
    suffix = '/share/private'
    key = 'monitor-base-url'
    setup_url_key = 'monitor-setup-url'

    t0 = t0 or time.time()
    base_url = cls.waitUntilPublished(instance_name, key, timeout, t0)
    setup_url = cls.getInstanceInfos(instance_name).connection_dict.get(
      setup_url_key)

    if not base_url or not setup_url:
      raise ValueError('Base URL or Setup URL is missing')

    # Parsing the setup URL to find the password
    parsed_setup_url = urllib.parse.urlparse(setup_url)
    query_params = urllib.parse.parse_qs(parsed_setup_url.query)
    password = query_params.get('password', [None])[0]
    if not password:
      raise ValueError('Password not found in the setup URL')

    # Adding credentials to the base URL
    parsed_base_url = urllib.parse.urlparse(base_url)
    netloc = f'admin:{password}@{parsed_base_url.hostname}'
    new_base_url = parsed_base_url._replace(netloc=netloc).geturl()
    final_url = f'{new_base_url}{suffix}'

    while True:
      resp = requests.get(final_url, verify=False)
      if resp.status_code == code:
        return resp, final_url

      msg = f'Instance {instance_name} monitor URL {final_url} returned unexpected status code {resp.status_code} (expected {code})'
      cls.checkTimeoutAndSleep(t0, timeout, msg)

  @classmethod
  def getMonitorPromises(cls, content):
    # Parse XML feed and extract test titles
    feed_xml = ET.fromstring(content)
    status = {}
    for item in feed_xml.findall("channel/item"):
      title = item.find("title").text.strip()
      description = item.find("description").text.strip()
      result = "[OK]" in title
      name = title.replace("[OK]", "").replace("[ERROR]", "").strip()
      status[name] = result
      cls.logger.info("Test alarm: %s: %s", title, description)
    return status

  @classmethod
  def waitUntilPromises(
      cls, instance_name, promise_name, expected, timeout=80, t0=None):
    t0 = t0 or time.time()
    msg = 'Instance %s promises not in expected state yet' % (instance_name)
    while True:
      resp, url = cls.waitUntilMonitorURLReady(
        instance_name, code=200, timeout=timeout, t0=t0)
      status = cls.getMonitorPromises(resp.content)
      cls.logger.info("Status: %s", status)
      cls.logger.info(
        "Promise Status: %s", status.get(promise_name, "Promise not found"))
      if status.get(promise_name) == expected:
        cls.logger.info(
          "%s is at expected status: %s" % (promise_name, expected))
        break
      cls.checkTimeoutAndSleep(t0, timeout, msg)



from slapos.tests.test_slapproxy import BasicMixin
import unittest
import httpretty

class HttpProxyTestCase(BasicMixin, unittest.TestCase):

  @httpretty.activate
  def test_full_request(self):
    url_to_proxy = 'http://example.org/my/path'
    httpretty.register_uri(
        httpretty.GET,
        url_to_proxy,
        status=201,
        body="bar\nfoo\n",
        adding_headers={
            # whilelisted
            "Content-Type": "foo1",
            "Date": "foo2",
            "Last-Modified": "foo3",
            "Vary": "foo4",
            "Cache-Control": "foo5",
            "Etag": "foo6",
            "Accept-Ranges": "foo7",
            "Content-Range": "foo8",
            "Www-Authenticate": "foo9",
            "Access-Control-Allow-Origin": "foo10",
            "Access-Control-Allow-Methods": "foo11",
            "Access-Control-Expose-Headers": "foo12",
            "Access-Control-Allow-Headers": "foo13",
            "Access-Control-Allow-Credentials": "foo14",
            # forced
            "Content-Disposition": "foobarfoo",
            # changed
            "Location": "http://example.org/my/new/path?a=b",
            # dropped
            "Set-Cookie": "Foo",
            "foo": "bar"
        }
    )
    response = self.app.get(
        '/http_proxy/http/example.org/my/path',
        query_string={
          "key1": "value1",
          "key2": "value2"
        },
        headers={
            # whitelisted
            "Content-Type": "bar1",
            "Accept": "bar2",
            "Accept-Language": "bar3",
            "Range": "bar4",
            "If-Modified-Since": "bar5",
            "If-None-Match": "bar6",
            "User-Agent": "bar7",
            "Authorization": "bar8",
            "Origin": "bar9",
            # rejected
            "Cookie": "bar10",
            "bar11": "bar11"
        },
        data="datafoo=databar"
    )

    # Response status code
    assert response.status_code == 201, response.status_code

    last_request = httpretty.last_request()

    # Request path and query string
    assert last_request.path in ['/my/path?key1=value1&key2=value2', '/my/path?key2=value2&key1=value1'], last_request.path

    # Request body
    assert last_request.body == b'datafoo=databar', last_request.body

    # Request headers propagated
    assert last_request.headers['Content-Type'] == 'bar1', last_request.headers
    assert last_request.headers['Accept'] == 'bar2', last_request.headers
    assert last_request.headers['Accept-Language'] == 'bar3', last_request.headers
    assert last_request.headers['Range'] == 'bar4', last_request.headers
    assert last_request.headers['If-Modified-Since'] == 'bar5', last_request.headers
    assert last_request.headers['If-None-Match'] == 'bar6', last_request.headers
    assert last_request.headers['User-Agent'] == 'bar7', last_request.headers
    assert last_request.headers['Authorization'] == 'bar8', last_request.headers
    assert last_request.headers['Origin'] == 'bar9', last_request.headers

    # Request headers blocked
    assert 'Cookie' not in last_request.headers, last_request.headers
    assert 'bar11' not in last_request.headers, last_request.headers

    # Response body
    assert response.data == b'bar\nfoo\n', response.data

    # Response header changed
    assert response.headers["Location"] == \
        "http://localhost/http_proxy/http/example.org/my/new/path?a=b", \
        response.headers

    # Response header forced
    assert response.headers["Content-Disposition"] == \
        "attachment", \
        response.headers

    # Response headers propagated
    assert response.headers["Content-Type"] == "foo1", response.headers
    assert response.headers["Date"] == "foo2", response.headers
    assert response.headers["Last-Modified"] == "foo3", response.headers
    assert response.headers["Vary"] == "foo4", response.headers
    assert response.headers["Cache-Control"] == "foo5", response.headers
    assert response.headers["Etag"] == "foo6", response.headers
    assert response.headers["Accept-Ranges"] == "foo7", response.headers
    assert response.headers["Content-Range"] == "foo8", response.headers
    assert response.headers["Www-Authenticate"] == "foo9", response.headers
    assert response.headers["Access-Control-Allow-Origin"] == "foo10", response.headers
    assert response.headers["Access-Control-Allow-Methods"] == "foo11", response.headers
    assert response.headers["Access-Control-Expose-Headers"] == "foo12", response.headers
    assert response.headers["Access-Control-Allow-Headers"] == "foo13", response.headers
    assert response.headers["Access-Control-Allow-Credentials"] == "foo14", response.headers

    # Response headers dropped
    assert 'Set-Cookie' not in response.headers, response.headers
    assert 'foo' not in response.headers, response.headers

  @httpretty.activate
  def test_path_propagated(self):
    httpretty.register_uri(
        httpretty.GET,
        'http://example.org/a/b',
        body="http://example.org/a/b",
    )
    httpretty.register_uri(
        httpretty.GET,
        'http://example.org/a',
        body="http://example.org/a",
    )
    httpretty.register_uri(
        httpretty.GET,
        'http://example.org/',
        body="http://example.org/",
    )
    httpretty.register_uri(
        httpretty.GET,
        'http://example.org',
        body="http://example.org",
    )

    response = self.app.get('/http_proxy/http/example.org')
    assert response.status_code == 200, response.status_code
    last_request = httpretty.last_request()
    assert last_request.path == '/', last_request.path
    assert response.data == b'http://example.org', response.data

    response = self.app.get('/http_proxy/http/example.org/')
    assert response.status_code == 200, response.status_code
    last_request = httpretty.last_request()
    assert last_request.path == '/', last_request.path
    assert response.data == b'http://example.org/', response.data

    response = self.app.get('/http_proxy/http/example.org/a')
    assert response.status_code == 200, response.status_code
    last_request = httpretty.last_request()
    assert last_request.path == '/a', last_request.path
    assert response.data == b'http://example.org/a', response.data

    response = self.app.get('/http_proxy/http/example.org/a/b')
    assert response.status_code == 200, response.status_code
    last_request = httpretty.last_request()
    assert last_request.path == '/a/b', last_request.path
    assert response.data == b'http://example.org/a/b', response.data



from werkzeug.http import HTTP_STATUS_CODES
from werkzeug.exceptions import NotAcceptable
from flask import current_app, request, abort, Blueprint, make_response, g, url_for
from .db import execute_db, requestInstanceFromDB, supplyFromDB, removeFromDB, \
                bangInstanceFromDB, getPartitionFromDB, freePartitionFromDB, \
                formatFromDB, \
                NotFoundPartitionFailure, PartitionDeletionFailure, \
                AllocationFailure, getInstanceTreeList, getAllocatedInstance
from slapos.util import dict2xml, xml2dict, loads
from slapos.slap.slap import ComputerPartition, SoftwareInstance
import json
import jsonschema
import sys


json_rpc_blueprint = Blueprint('json_rpc', __name__)
json_rpc_experimental_blueprint = Blueprint('json_rpc_experimental', __name__)


def find_best_match(status_code=None):
  if (status_code == 406) or (len(request.accept_mimetypes) == 0):
    mime = 'application/json'
  else:
    mime = request.accept_mimetypes.best_match(['application/json'])
    if mime is None:
      # Didn't find good mimetype.
      raise NotAcceptable
  return mime


def send_json_rpc_document(json_rpc_dict, status_code=None, mime=None):
  if mime is None:
    try:
      mime = find_best_match(status_code=status_code)
    except NotAcceptable:
      return current_app.handle_http_exception(NotAcceptable)

  data = json.dumps(json_rpc_dict)
  if status_code:
    rv = make_response((data, status_code))
  else:
    rv = make_response(data)

  rv.headers['Content-Type'] = mime
  return rv


def validate_and_send_json_rpc_document(json_rpc_dict, **kw):
  output_schema_text = g.output_schema_text
  # validate output
  try:
    jsonschema.validate(
      json_rpc_dict,
      output_schema_text,
      format_checker=jsonschema.FormatChecker()
    )
  except jsonschema.exceptions.ValidationError as err:
    return abort(500, err.message)

  return send_json_rpc_document(json_rpc_dict, **kw)


def before_request(open_api_json_file_name):
  if (request.method != "POST"):
    return abort(405)

  # Validate the input json
  with json_rpc_blueprint.open_resource(open_api_json_file_name, 'r') as json_rpc_open_api_file:
    config_filejson_rpc_parsed_json = json.loads(json_rpc_open_api_file.read())

  try:
    method = config_filejson_rpc_parsed_json['paths'][request.path]['post']
  except KeyError:
    return abort(500, '%s is not part of the open api definition' % request.path)

  input_schema_text = method['requestBody']['content']['application/json']['schema']
  g.output_schema_text = method['responses']['200']['content']['application/json']['schema']

  # Validate the input body
  body_json = request.json
  try:
    jsonschema.validate(
      body_json,
      input_schema_text,
      format_checker=jsonschema.FormatChecker()
    )
  except jsonschema.exceptions.ValidationError as err:
    return abort(400, err.message)

def before_json_rpc_request():
  return before_request('json_rpc.json')

def before_json_rpc_experimental_request():
  return before_request('json_rpc_experimental.json')

json_rpc_blueprint.before_request(before_json_rpc_request)
json_rpc_experimental_blueprint.before_request(before_json_rpc_experimental_request)


class JsonRpcManager(object):
  '''
  This object is used to replicate the slapos master json rpc api
  '''
  def init_app(self, app, **kw):
    app.register_blueprint(json_rpc_blueprint, **kw)
    app.register_blueprint(json_rpc_experimental_blueprint, **kw)
    app.handle_exception = self._custom_handle_exception
    app.handle_http_exception = self._custom_handle_http_exception

  # Redefine http exception handling to return JSON
  def _custom_handle_http_exception(self, exception):
    if request.blueprint in ['json_rpc', 'json_rpc_experimental']:
      error_dict = {
        "status": exception.code,
        "type": HTTP_STATUS_CODES.get(exception.code, 'Unknown Error'),
        "title": exception.description
      }
      return send_json_rpc_document(error_dict, status_code=exception.code)
    return exception

  # Redefine python exception handling to return JSON
  def _custom_handle_exception(self, exception):
    if request.blueprint in ['json_rpc', 'json_rpc_experimental']:
      # Log exception and return json
      exc_type, exc_value, tb = sys.exc_info()
      current_app.log_exception((exc_type, exc_value, tb))
      return send_json_rpc_document({
        'status': 500,
        'type': HTTP_STATUS_CODES.get(500, 'Unknown Error'),
        'title': 'Internal server error'
      }, status_code=500)
    raise exception


@json_rpc_blueprint.route('/slapos.remove.v0.compute_node_certificate', methods=['POST'])
def remove_compute_node_certificate():
  return validate_and_send_json_rpc_document({
    'type': 'success',
    'title': 'Ignored'
  })

@json_rpc_blueprint.route('/slapos.post.v0.software_installation', methods=['POST'])
def supply_software_installation():
  computer_id = request.json["computer_guid"]
  software_release_uri = request.json["software_release_uri"]
  state = request.json.get("state", "available")
  supplyFromDB(computer_id, software_release_uri, state)
  return validate_and_send_json_rpc_document({
    'type': 'success',
    'title': 'Supplied %s to be %s' % (software_release_uri, state)
  })

@json_rpc_blueprint.route('/slapos.post.v0.compute_node_usage', methods=['POST'])
def post_compute_node_usage():
  return validate_and_send_json_rpc_document({
    'type': 'success',
    'title': 'Ignored'
  })

@json_rpc_blueprint.route('/slapos.post.v0.compute_node_certificate', methods=['POST'])
def post_compute_node_certificate():
  # proxy does not use node certificate, but client calls this.
  return validate_and_send_json_rpc_document({
    'key': '',
    'certificate': ''
  })

@json_rpc_blueprint.route('/slapos.allDocs.v0.compute_node_software_installation_list', methods=['POST'])
def compute_node_software_installation_list():
  computer_id = request.json["computer_guid"]
  computer_list = execute_db('computer', 'SELECT * FROM %s WHERE reference=?', [computer_id])
  if len(computer_list) != 1:
    return abort(403, '%s is not registered.' % computer_id)
  software_release_list = []
  for sr in execute_db('software', 'select * from %s WHERE computer_reference=?', [computer_id]):
    software_release_list.append({
      'software_release_uri': sr['url'],
      'state': sr['requested_state']
    })
  return validate_and_send_json_rpc_document({
    'result_list': software_release_list
  })

def generateInstanceGuid(title, requested_by, is_shared):
  # We expect slapproxy to only a uniq partition_reference/requested_by
  return '%s___%s___%i' % (title, requested_by, int(is_shared))

def extractInstanceGuid(instance_guid):
  result_list = instance_guid.split('___', 2)
  result_list[2] = bool(int(result_list[2]))
  return result_list

def generateDBSlaveReference(title, requested_by):
  return '%s_%s' % (requested_by, title)

def extractDBSlaveReference(reference):
  requested_by, title = reference.split('_', 1)
  return title, requested_by


@json_rpc_blueprint.route('/slapos.allDocs.v0.compute_node_instance_list', methods=['POST'])
def compute_node_instance_list():
  computer_id = request.json["computer_guid"]
  computer_list = execute_db('computer', 'SELECT * FROM %s WHERE reference=?', [computer_id])
  if len(computer_list) != 1:
    return abort(403, '%s is not registered.' % computer_id)
  instance_list = []
  for partition in execute_db('partition', 'SELECT * FROM %s WHERE computer_reference=? AND slap_state="busy"', [computer_id]):
    instance_list.append({
      "title": partition['partition_reference'],
      "instance_guid": generateInstanceGuid(partition['partition_reference'], partition['requested_by'], False),
      "state": partition['requested_state'],
      "compute_partition_id": partition['reference'],
      "software_release_uri": partition['software_release'],
    })
  return validate_and_send_json_rpc_document({
    'result_list': instance_list
  })

@json_rpc_blueprint.route('/slapos.allDocs.v0.instance_node_instance_list', methods=['POST'])
def instance_node_instance_list():
  try:
    partition_reference, requested_by, is_shared = extractInstanceGuid(request.json["instance_guid"])
  except (ValueError, IndexError):
    return abort(403, 'instance_guid %s not handled.' % request.json["instance_guid"])

  partition = getAllocatedInstance(partition_reference, requested_by)
  if is_shared or (not partition):
    return abort(403, 'No software instance %s found.' % request.json["instance_guid"])

  result_list = []
  slave_instance_list = []
  if partition['slave_instance_list'] is not None:
    slave_instance_list = loads(partition['slave_instance_list'].encode('utf-8'))
  for slave_information in slave_instance_list:
    title, asked_by = extractDBSlaveReference(slave_information.pop('slave_title'))
    # Drop added informations
    slap_software_type = slave_information.pop('slap_software_type')
    slave_information.pop('slave_reference')
    result_list.append({
      "title": title,
      "instance_guid": generateInstanceGuid(title, asked_by, True),
      "software_type": slap_software_type,
      # The state is not stored in slapproxy
      "state": "started",
      "parameters": slave_information,
      "compute_partition_id": partition['reference']
    })
  return validate_and_send_json_rpc_document({
    'result_list': result_list
  })

@json_rpc_blueprint.route('/slapos.get.v0.hateoas_url', methods=['POST'])
def get_hateoas_url():
  return validate_and_send_json_rpc_document({
    'hateoas_url': url_for('hateoas.hateoas', _external=True)
  })

@json_rpc_blueprint.route('/slapos.get.v0.compute_node_status', methods=['POST'])
def get_compute_node_status():
  return validate_and_send_json_rpc_document({
    'text': 'Unknown (not implemented)'
  })

@json_rpc_blueprint.route('/slapos.get.v0.software_instance_certificate', methods=['POST'])
def get_software_instance_certificate():
  # proxy does not use partition certificate, but client calls this.
  return validate_and_send_json_rpc_document({
    'key': '',
    'certificate': ''
  })

def send_json_rpc_sql_partition(partition):
  address_list = []
  for address in execute_db('partition_network',
                            'SELECT * FROM %s WHERE partition_reference=? AND computer_reference=?',
                            (partition['reference'], partition['computer_reference'])):
    address_list.append([address['reference'], address['address']])

  return validate_and_send_json_rpc_document({
    "title": partition['partition_reference'],
    "instance_guid": generateInstanceGuid(partition['partition_reference'], partition['requested_by'], False),
    "software_release_uri": partition['software_release'],
    "software_type": partition['software_type'],
    "state": partition['requested_state'],
    "connection_parameters": xml2dict(partition['connection_xml']),
    "parameters": xml2dict(partition['xml']),
    "shared": False,
    "root_instance_title": partition['requested_by'] or partition['partition_reference'],
    "ip_list": address_list,
    "full_ip_list": [],
    # sla are not stored in slapproxy
    "sla_parameters": {},
    "computer_guid": partition['computer_reference'],
    "compute_partition_id": partition['reference'],
    "processing_timestamp": int(partition['timestamp'] or '0'),
    # This info is probably not available
    "access_status_message": ''
  })


def send_json_rpc_sql_shared(shared_instance, partition):
  title, asked_by = extractDBSlaveReference(shared_instance['reference'])
  assert asked_by == shared_instance['asked_by']
  slave_information = [x for x in loads(partition['slave_instance_list'].encode('utf-8')) if x['slave_title'] == shared_instance['reference']][0]
  # Drop added informations
  slave_information.pop('slave_title')
  slave_information.pop('slap_software_type')
  slave_information.pop('slave_reference')

  return validate_and_send_json_rpc_document({
    "title": title,
    "instance_guid": generateInstanceGuid(title, shared_instance['asked_by'], True),
    "software_release_uri": partition['software_release'],
    "software_type": partition['software_type'],
    # XXX This is not stored
    "state": "started",
    "connection_parameters": xml2dict(shared_instance['connection_xml']),
    "parameters": slave_information,
    "shared": True,
    "root_instance_title": shared_instance['asked_by'] or title,
    "ip_list": [],
    "full_ip_list": [],
    # sla are not stored in slapproxy
    "sla_parameters": {},
    "computer_guid": partition['computer_reference'],
    "compute_partition_id": partition['reference'],
    "processing_timestamp": int(partition['timestamp']),
    # This info is probably not available
    "access_status_message": ''
  })

def send_json_rpc_slap_instance(title, requested_by, is_shared, slap_instance):
  parameters = slap_instance._parameter_dict
  if is_shared:
    # XXX there is not timestamp for shared instance currently
    timestamp = 0
    state = 'started'
  else:
    timestamp = parameters.pop('timestamp')
    state = slap_instance._requested_state
  return validate_and_send_json_rpc_document({
    "title": title,
    "instance_guid": generateInstanceGuid(title, requested_by, is_shared),
    "software_release_uri": slap_instance.slap_software_release_url,
    "software_type": slap_instance.slap_software_type,
    "state": state,
    "connection_parameters": slap_instance._connection_dict,
    "parameters": parameters,
    "shared": is_shared,
    "root_instance_title": requested_by or title,
    "ip_list": [[y for y in x] for x in slap_instance.ip_list],
    "full_ip_list": [],
    # sla are not stored in slapproxy
    "sla_parameters": {},
    "computer_guid": slap_instance.slap_computer_id,
    "compute_partition_id": slap_instance.slap_computer_partition_id,
    "processing_timestamp": int(float(timestamp)),
    # This info is probably not available
    "access_status_message": ''
  })

@json_rpc_blueprint.route('/slapos.get.v0.software_instance', methods=['POST'])
def get_software_instance():
  try:
    partition_reference, requested_by, is_shared = extractInstanceGuid(request.json["instance_guid"])
  except (ValueError, IndexError):
    return abort(403, 'instance_guid %s not handled.' % request.json["instance_guid"])

  if is_shared:
    # XXX Copied from db.getAllocatedSlaveInstance
    shared = execute_db('slave',
      'SELECT * FROM %s WHERE reference=? AND asked_by=?',
      (generateDBSlaveReference(partition_reference, requested_by), requested_by), one=True)
    if not shared:
      return abort(403, 'No shared instance %s found.' % request.json["instance_guid"])
    # Get the matching partition
    partition = execute_db('partition',
      'SELECT * FROM %s WHERE computer_reference=? AND reference=?',
      (shared['computer_reference'], shared['hosted_by']), one=True)
    if not partition:
      return abort(403, 'No partition hosting %s found.' % request.json["instance_guid"])
    return send_json_rpc_sql_shared(shared, partition)

  partition = getAllocatedInstance(partition_reference, requested_by)
  if not partition:
    return abort(403, 'No software instance %s found.' % request.json["instance_guid"])
  return send_json_rpc_sql_partition(partition)

@json_rpc_blueprint.route('/slapos.get.v0.compute_partition', methods=['POST'])
def get_compute_partition():
  partition = getPartitionFromDB(request.json["compute_partition_id"], request.json["computer_guid"])
  if (not partition) or (not partition['partition_reference']):
    return abort(403, 'No instance on partition %s found.' % request.json["compute_partition_id"])
  return send_json_rpc_sql_partition(partition)

@json_rpc_blueprint.route('/slapos.post.v0.software_instance', methods=['POST'])
def post_software_instance():
  title = request.json["title"]
  requested_by = ''# XXX getRequesterFromForm(form) or '',
  parameters = request.json.get("parameters", {})
  is_shared = request.json.get("shared", False)
  requested_state = request.json.get("state", "started")
  parsed_request_dict = {
    'requester_id': None,
    'requested_by': requested_by,
    'software_release': request.json["software_release_uri"],
    'software_type': request.json["software_type"],
    'partition_reference': title,
    'partition_parameter_kw': parameters,
    'filter_kw': request.json.get("sla_parameters", {}),
    # Note: currently ignored for slave instance (slave instances
    # are always started).
    'requested_state': requested_state,
    # Is it a slave instance?
    'slave': is_shared
  }
  try:
    slap_instance = requestInstanceFromDB(**parsed_request_dict)
  except AllocationFailure as e:
    return abort(403, str(e))
  if isinstance(slap_instance, SoftwareInstance):
    return send_json_rpc_slap_instance(title, requested_by, is_shared, slap_instance)

  elif isinstance(slap_instance, ComputerPartition):
    return validate_and_send_json_rpc_document({
      "title": title,
      "instance_guid": generateInstanceGuid(title, requested_by, is_shared),
      "software_release_uri": request.json["software_release_uri"],
      "software_type": request.json["software_type"],
      "state": requested_state,
      "connection_parameters": slap_instance._connection_dict,
      "parameters": parameters,
      "shared": is_shared,
      "root_instance_title": requested_by or title,
      "ip_list": [],
      "full_ip_list": [],
      # sla are not stored in slapproxy
      "sla_parameters": {},
      "computer_guid": slap_instance._computer_id,
      "compute_partition_id": slap_instance._partition_id,
      "processing_timestamp": 0,
      # This info is probably not available
      "access_status_message": ''
    })

  return abort(500, 'Can not export %s' % str(slap_instance))

@json_rpc_blueprint.route('/slapos.put.v0.compute_node_format', methods=['POST'])
def put_compute_node_format():
  formatFromDB(request.json["computer_guid"],
               request.json["compute_partition_list"])
  return validate_and_send_json_rpc_document({
    'type': 'success',
    'title': 'Formatted'
  })

@json_rpc_blueprint.route('/slapos.put.v0.compute_node_bang', methods=['POST'])
def put_compute_node_bang():
  return validate_and_send_json_rpc_document({
    'type': 'success',
    'title': 'Ignored'
  })

@json_rpc_blueprint.route('/slapos.put.v0.software_installation_reported_state', methods=['POST'])
def put_software_installation_reported_state():
  reported_state = request.json["reported_state"]
  if reported_state in ["available", "building"]:
    return validate_and_send_json_rpc_document({
      'type': 'success',
      'title': 'Ignored'
    })
  elif reported_state == "destroyed":
    removeFromDB(request.json["computer_guid"], request.json["software_release_uri"])
    return validate_and_send_json_rpc_document({
      'type': 'success',
      'title': 'Destroyed'
    })
  else:
    raise NotImplementedError('State %s handling not implemented' % reported_state)

@json_rpc_blueprint.route('/slapos.put.v0.software_instance_reported_state', methods=['POST'])
def put_software_instance_reported_state():
  reported_state = request.json["reported_state"]
  if reported_state in ['started', 'stopped']:
    return validate_and_send_json_rpc_document({
      'type': 'success',
      'title': 'Ignored'
    })

  if reported_state == 'destroyed':
    try:
      partition_reference, requested_by, is_shared = extractInstanceGuid(request.json["instance_guid"])
    except (ValueError, IndexError):
      return abort(403, 'instance_guid %s not handled.' % request.json["instance_guid"])

    if is_shared:
      return validate_and_send_json_rpc_document({
        'type': 'success',
        'title': 'Ignored'
      })
    else:

      # Instance is fetched twice from DB
      # Far from optimal, but enough for now, as this is really
      # not called often
      partition = getAllocatedInstance(partition_reference, requested_by)
      if not partition:
        return abort(403, 'No software instance %s found.' % request.json["instance_guid"])

      try:
        freePartitionFromDB(partition['reference'], partition['computer_reference'])
      except (NotFoundPartitionFailure, PartitionDeletionFailure) as error:
        return abort(403, str(error))

      return validate_and_send_json_rpc_document({
        'type': 'success',
        'title': 'Destroyed'
      })

  else:
    raise NotImplementedError('State %s handling not implemented' % reported_state)

@json_rpc_blueprint.route('/slapos.put.v0.software_instance_bang', methods=['POST'])
def put_software_instance_bang():
  try:
    partition_reference, requested_by, is_shared = extractInstanceGuid(request.json["instance_guid"])
  except (ValueError, IndexError):
    return abort(403, 'instance_guid %s not handled.' % request.json["instance_guid"])

  if is_shared:
    return abort(403, 'NotImplemented')
  else:
    bangInstanceFromDB(partition_reference, requested_by)

  return validate_and_send_json_rpc_document({
    'type': 'success',
    'title': 'Bang handled'
  })

@json_rpc_blueprint.route('/slapos.put.v0.software_instance_title', methods=['POST'])
def put_software_instance_title():
  try:
    partition_reference, requested_by, is_shared = extractInstanceGuid(request.json["instance_guid"])
  except (ValueError, IndexError):
    return abort(403, 'instance_guid %s not handled.' % request.json["instance_guid"])
  new_title = request.json['title']

  if is_shared:
    return abort(403, 'NotImplemented')
    """
    # XXX TODO update slave_instance_list
    query = 'UPDATE %s SET reference=? WHERE reference=? AND asked_by=?'
    argument_list = [
      generateDBSlaveReference(new_title, requested_by),
      generateDBSlaveReference(partition_reference, requested_by),
      requested_by
    ]
    execute_db('slave', query, argument_list)
    """
  else:
    query = 'UPDATE %s SET partition_reference=? WHERE partition_reference=? AND requested_by=?'
    argument_list = [new_title, partition_reference, requested_by]
    execute_db('partition', query, argument_list)

  return validate_and_send_json_rpc_document({
    'type': 'success',
    'title': 'Renamed'
  })

@json_rpc_blueprint.route('/slapos.put.v0.software_instance_connection_parameter', methods=['POST'])
def put_software_instance_connection_parameter():
  try:
    partition_reference, requested_by, is_shared = extractInstanceGuid(request.json["instance_guid"])
  except (ValueError, IndexError):
    return abort(403, 'instance_guid %s not handled.' % request.json["instance_guid"])
  connection_xml = dict2xml(request.json["connection_parameter_dict"])

  if is_shared:
    query = 'UPDATE %s SET connection_xml=? WHERE reference=? AND asked_by=?'
    argument_list = [connection_xml, generateDBSlaveReference(partition_reference, requested_by), requested_by]
    execute_db('slave', query, argument_list)
  else:
    query = 'UPDATE %s SET connection_xml=? WHERE partition_reference=? AND requested_by=?'
    argument_list = [connection_xml, partition_reference, requested_by]
    execute_db('partition', query, argument_list)

  return validate_and_send_json_rpc_document({
    'type': 'success',
    'title': 'Updated'
  })


@json_rpc_blueprint.route('/slapos.put.v0.software_instance_error', methods=['POST'])
def put_software_instance_error():
  return validate_and_send_json_rpc_document({
    'type': 'success',
    'title': 'Ignored'
  })

@json_rpc_blueprint.route('/slapos.put.v0.software_installation_error', methods=['POST'])
def put_software_installation_error():
  return validate_and_send_json_rpc_document({
    'type': 'success',
    'title': 'Ignored'
  })


@json_rpc_experimental_blueprint.route('/slapos.allDocs.WIP.instance_tree_list', methods=['POST'])
def instance_tree_list():
  partition_and_shared_list = getInstanceTreeList()
  result_list = []
  for partition in partition_and_shared_list[0]:
    result_list.append({
      "instance_guid": generateInstanceGuid(partition['partition_reference'], '', False),
      "title": partition['partition_reference']
    })
  for shared in partition_and_shared_list[1]:
    title, asked_by = extractDBSlaveReference(shared['reference'])
    assert asked_by == shared['asked_by']
    result_list.append({
      "instance_guid": generateInstanceGuid(title, '', True),
      "title": title
    })
  return validate_and_send_json_rpc_document({
    'result_list': result_list
  })

@json_rpc_experimental_blueprint.route('/slapos.allDocs.WIP.compute_node_list', methods=['POST'])
def compute_node_list():
  result_list = []
  for computer in execute_db('computer', 'SELECT reference FROM %s', ()):
    result_list.append({
      "computer_guid": computer['reference']
    })
  return validate_and_send_json_rpc_document({
    'result_list': result_list
  })



##############################################################################
#
# Copyright (c) 2014 Vifib SARL and Contributors. All Rights Reserved.
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsibility of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# guarantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

from slapos.util import mkdir_p
from datetime import datetime, timedelta
import csv
import six
import mock
import json
import time
import os
import glob
import unittest
import shutil
import tarfile
import tempfile
import slapos.slap
import psutil
import sqlite3
import subprocess
from slapos.collect import entity, snapshot, db, reporter
from slapos.cli.entry import SlapOSApp
from six.moves.configparser import ConfigParser

from lxml import etree as ElementTree
import slapos.tests.data


class FakeDatabase(object):
    def __init__(self):
      self.invoked_method_list = []

    def connect(self):
      self.invoked_method_list.append(("connect", ""))

    def close(self):
      self.invoked_method_list.append(("close", ""))

    def commit(self):
      self.invoked_method_list.append(("commit", ""))

    def insertUserSnapshot(self, *args, **kw):
      self.invoked_method_list.append(("insertUserSnapshot", (args, kw)))

    def inserFolderSnapshot(self, *args, **kw):
      self.invoked_method_list.append(("inserFolderSnapshot", (args, kw)))

    def insertSystemSnapshot(self, *args, **kw):
      self.invoked_method_list.append(("insertSystemSnapshot", (args, kw)))

    def insertComputerSnapshot(self, *args, **kw):
      self.invoked_method_list.append(("insertComputerSnapshot", (args, kw)))

    def insertDiskPartitionSnapshot(self, *args, **kw):
      self.invoked_method_list.append(("insertDiskPartitionSnapshot", (args, kw)))

    def insertTemperatureSnapshot(self, *args, **kw):
      self.invoked_method_list.append(("insertTemperatureSnapshot", (args, kw)))

    def insertHeatingSnapshot(self, *args, **kw):
      self.invoked_method_list.append(("insertHeatingSnapshot", (args, kw)))

class FakeDatabase2(FakeDatabase):
    def select(self, *args, **kw):
      self.invoked_method_list.append(("select", (args, kw)))
      return []

class TestCollectDatabase(unittest.TestCase):

    def setUp(self):
        self.instance_root = tempfile.mkdtemp()

    def tearDown(self):
        if os.path.exists(self.instance_root):
          shutil.rmtree(self.instance_root)

    def test_database_bootstrap(self):
        self.assertFalse(os.path.exists(
                  "%s/collector.db" % self.instance_root ))
        database = db.Database(self.instance_root, create=True)
        database.connect()
        try:
          self.assertEqual(
              [u'user', u'folder', u'computer', u'system', u'disk', u'temperature', u'heating'],
              database.getTableList())
        finally:
          database.close()

        self.assertTrue(os.path.exists(
                  "%s/collector.db" % self.instance_root ))

    def test_database_not_bootstrap(self):
        self.assertFalse(os.path.exists(
                  "%s/collector.db" % self.instance_root ))
        database = db.Database(self.instance_root)
        database.connect()
        try:
          self.assertNotEqual(
              [u'user', u'folder', u'computer', u'system', u'disk', u'temperature', u'heating'],
              database.getTableList())
        finally:
          database.close()

        self.assertTrue(os.path.exists(
                  "%s/collector.db" % self.instance_root ))

    def test_database_select(self):
      def _fake_execute(sql): return sql

      database = db.Database(self.instance_root, create=True)
      database.connect()
      original_execute = database._execute
      try:
        database._execute = _fake_execute
        self.assertEqual("SELECT * FROM user  ", database.select('user'))
        self.assertEqual("SELECT DATE FROM user  WHERE date = '0.1'  AND time=\"00:02\"  ",
                          database.select('user', 0.1, 'DATE', 'time="00:02"'))
        self.assertEqual("SELECT DATE FROM user  WHERE date = '0.1'   GROUP BY time ORDER BY date",
                          database.select('user', 0.1, 'DATE', order='date', group='time' ))
        self.assertEqual("SELECT DATE FROM user  WHERE date = '0.1'   limit 1",
                          database.select('user', 0.1, 'DATE', limit=1))
      finally:
        database._execute = original_execute
        database.close()

    def test_insert_user_snapshot(self):

        database = db.Database(self.instance_root, create=True)
        database.connect()
        try:
            database.insertUserSnapshot(
             'fakeuser0', 10, '10-12345', '0.1', '10.0', '1',
             '10.0', '10.0', '0.1', '0.1', 'DATE', 'TIME')
            database.commit()
            self.assertEqual([i for i in database.select('user')], 
                          [(u'fakeuser0', 10.0, u'10-12345', 0.1, 10.0, 
                          1.0, 10.0, 10.0, 0.1, 0.1, u'DATE', u'TIME', 0)])
        finally:
            database.close()

    def test_insert_folder_snapshot(self):

        database = db.Database(self.instance_root, create=True)
        database.connect()
        try:
            database.inserFolderSnapshot(
             'fakeuser0', '0.1', 'DATE', 'TIME')
            database.commit()
            self.assertEqual([i for i in database.select('folder')], 
                          [(u'fakeuser0', 0.1, u'DATE', u'TIME', 0)])
        finally:
            database.close()

    def test_insert_computer_snapshot(self):

        database = db.Database(self.instance_root, create=True)
        database.connect()
        try:
            database.insertComputerSnapshot(
              '1', '0', '0', '100', '0', '/dev/sdx1', 'DATE', 'TIME')
            database.commit()
            self.assertEqual([i for i in database.select('computer')], 
                    [(1.0, 0.0, u'0', 100.0, u'0', u'/dev/sdx1', u'DATE', u'TIME', 0)]) 
        finally:
          database.close()

    def test_insert_disk_partition_snapshot(self):

        database = db.Database(self.instance_root, create=True)
        database.connect()
        try:
            database.insertDiskPartitionSnapshot(
                 '/dev/sdx1', '10', '20', '/mnt', 'DATE', 'TIME')
            database.commit() 
            self.assertEqual([i for i in database.select('disk')], 
                            [(u'/dev/sdx1', u'10', u'20', u'/mnt', u'DATE', u'TIME', 0)])
        finally:
          database.close()

    def test_insert_system_snapshot(self):

        database = db.Database(self.instance_root, create=True)
        database.connect()
        try:
            database.insertSystemSnapshot("0.1", '10.0', '100.0', '100.0', 
                         '10.0', '1', '2', '12.0', '1', '1', 'DATE', 'TIME')
            database.commit()

            self.assertEqual([i for i in database.select('system')], 
                             [(0.1, 10.0, 100.0, 100.0, 10.0, 1.0, 
                               2.0, 12.0, 1.0, 1.0, u'DATE', u'TIME', 0)])
        finally:
          database.close()

    def test_date_scope(self):

        database = db.Database(self.instance_root, create=True)
        database.connect()
        try:
            database.insertSystemSnapshot("0.1", '10.0', '100.0', '100.0', 
                 '10.0', '1', '2', '12.0', '1', '1', 'EXPECTED-DATE', 'TIME')
            database.commit()

            self.assertEqual([i for i in database.getDateScopeList()], 
                             [('EXPECTED-DATE', 1)])

            self.assertEqual([i for i in \
               database.getDateScopeList(ignore_date='EXPECTED-DATE')], 
               [])

            self.assertEqual([i for i in \
               database.getDateScopeList(reported=1)], 
               [])

        finally:
          database.close()

    def test_garbage_collection_date_list(self):
        database = db.Database(self.instance_root, create=True)
        self.assertEqual(len(database._getGarbageCollectionDateList(3)), 3)
        self.assertEqual(len(database._getGarbageCollectionDateList(1)), 1)
        self.assertEqual(len(database._getGarbageCollectionDateList(0)), 0)

        self.assertEqual(database._getGarbageCollectionDateList(1), 
                           [datetime.utcnow().strftime("%Y-%m-%d")])

    def test_garbage(self):

        database = db.Database(self.instance_root, create=True)
        database.connect()
        database.insertSystemSnapshot("0.1", '10.0', '100.0', '100.0', 
                         '10.0', '1', '2', '12.0', '1', '1', '1983-01-10', 'TIME')
        database.insertDiskPartitionSnapshot(
                 '/dev/sdx1', '10', '20', '/mnt', '1983-01-10', 'TIME')
        database.insertComputerSnapshot(
              '1', '0', '0', '100', '0', '/dev/sdx1', '1983-01-10', 'TIME')
        database.insertUserSnapshot(
             'fakeuser0', 10, '10-12345', '0.1', '10.0', '1',
             '10.0', '10.0', '0.1', '0.1', '1983-01-10', 'TIME')
        database.inserFolderSnapshot(
             'fakeuser0', '0.1', '1983-01-10', 'TIME')
        database.commit()
        database.markDayAsReported(date_scope="1983-01-10", 
                                       table_list=database.getTableList())
        database.commit()
        self.assertEqual(len([x for x in database.select('system')]), 1)
        self.assertEqual(len([x for x in database.select('folder')]), 1)
        self.assertEqual(len([x for x in database.select('user')]), 1)
        #self.assertEqual(len([x for x in database.select('heating')]), 1)
        #self.assertEqual(len([x for x in database.select('temperature')]), 1)
        self.assertEqual(len([x for x in database.select('computer')]), 1)
        self.assertEqual(len([x for x in database.select('disk')]), 1)
        database.close()

        database.garbageCollect()
        database.connect()
        self.assertEqual(len([x for x in database.select('folder')]), 0)
        self.assertEqual(len([x for x in database.select('user')]), 0)
        #self.assertEqual(len([x for x in database.select('heating')]), 0)
        #self.assertEqual(len([x for x in database.select('temperature')]), 0)
        self.assertEqual(len([x for x in database.select('system')]), 0)
        self.assertEqual(len([x for x in database.select('computer')]), 0)
        self.assertEqual(len([x for x in database.select('disk')]), 0)

    def test_mark_day_as_reported(self):

        database = db.Database(self.instance_root, create=True)
        database.connect()
        try:
            database.insertSystemSnapshot("0.1", '10.0', '100.0', '100.0', 
                 '10.0', '1', '2', '12.0', '1', '1', 'EXPECTED-DATE', 'TIME')
            database.insertSystemSnapshot("0.1", '10.0', '100.0', '100.0', 
                 '10.0', '1', '2', '12.0', '1', '1', 'NOT-EXPECTED-DATE', 'TIME')
            database.commit()

            self.assertEqual([i for i in database.select('system')], 
                             [(0.1, 10.0, 100.0, 100.0, 10.0, 1.0, 
                               2.0, 12.0, 1.0, 1.0, u'EXPECTED-DATE', u'TIME', 0),
                             (0.1, 10.0, 100.0, 100.0, 10.0, 1.0, 
                               2.0, 12.0, 1.0, 1.0, u'NOT-EXPECTED-DATE', u'TIME', 0)])

            database.markDayAsReported(date_scope="EXPECTED-DATE", 
                                       table_list=["system"])
            database.commit()

            self.assertEqual([i for i in database.select('system')], 
                             [(0.1, 10.0, 100.0, 100.0, 10.0, 1.0, 
                               2.0, 12.0, 1.0, 1.0, u'EXPECTED-DATE', u'TIME', 1),
                             (0.1, 10.0, 100.0, 100.0, 10.0, 1.0, 
                               2.0, 12.0, 1.0, 1.0, u'NOT-EXPECTED-DATE', u'TIME', 0)])

        finally:
          database.close()

class TestCollectReport(unittest.TestCase):

    def setUp(self):
        self.instance_root = tempfile.mkdtemp()

    def tearDown(self):
        if os.path.exists(self.instance_root):
          shutil.rmtree(self.instance_root)

    def getPopulatedDB(self, day='1983-01-10', amount=1):
        database = db.Database(self.instance_root, create=True)
        database.connect()
        for i in range(0, amount):
          database.insertSystemSnapshot("0.1", '10.0', '100.0', '100.0', 
                           '10.0', '1', '2', '12.0', '1', '1', day, 'TIME')
          database.insertDiskPartitionSnapshot(
                   '/dev/sdx1', '10', '20', '/mnt', day, 'TIME')
          database.insertComputerSnapshot(
                '1', '0', '0', '100', '0', '/dev/sdx1', day, 'TIME')
          database.insertUserSnapshot(
               'fakeuser0', 10, '10-12345', '0.1', '10.0', '1',
               '10.0', '10.0', '0.1', '0.1', day, 'TIME')
          database.inserFolderSnapshot(
               'fakeuser0', '0.1', day, 'TIME')
        database.commit()
        database.close()
        return database

    def _get_file_content(self, f_path):
        with open(f_path, "r") as f:
          return f.readlines()

    def test_raw_csv_report(self):

        database = self.getPopulatedDB(amount=1)
        reporter.RawCSVDumper(database).dump(self.instance_root)
        self.assertTrue(os.path.exists("%s/1983-01-10" % self.instance_root))

        csv_path_dict = {
          '%s/1983-01-10/dump_disk.csv' % self.instance_root : 
            [['/dev/sdx1','10','20','/mnt','1983-01-10','TIME','0']],
          '%s/1983-01-10/dump_computer.csv' % self.instance_root :
            [['1.0','0.0','0','100.0','0','/dev/sdx1','1983-01-10','TIME','0']],
          '%s/1983-01-10/dump_user.csv' % self.instance_root :
            [['fakeuser0','10.0','10-12345','0.1','10.0','1.0','10.0','10.0','0.1','0.1','1983-01-10','TIME','0']],
          '%s/1983-01-10/dump_folder.csv' % self.instance_root :
            [['fakeuser0','0.1','1983-01-10','TIME','0']],
          '%s/1983-01-10/dump_heating.csv' % self.instance_root : [],
          '%s/1983-01-10/dump_temperature.csv' % self.instance_root : [],
          '%s/1983-01-10/dump_system.csv' % self.instance_root :
            [['0.1','10.0','100.0','100.0','10.0','1.0','2.0','12.0','1.0','1.0','1983-01-10','TIME','0']]}

        self.assertEqual(set(glob.glob("%s/1983-01-10/*.csv" % self.instance_root)),
                          set(csv_path_dict.keys()))

        
        for f_path in list(set(glob.glob("%s/1983-01-10/*.csv" % self.instance_root))):
          with open(f_path, "r") as csv_file:
            self.assertEqual([i for i in csv.reader(csv_file)], csv_path_dict[f_path])

    def test_system_json_report(self):
        database = self.getPopulatedDB(datetime.utcnow().strftime("%Y-%m-%d"), amount=2)
        reporter.SystemJSONReporterDumper(database).dump(self.instance_root)
        date_to_test = datetime.utcnow().strftime("%Y-%m-%d")
        json_path_dict = {
          '%s/system_memory_used.json' % self.instance_root: 
            '[  {    "entry": 0.09765625,     "time": "%s TIME"  }]' % date_to_test,
          '%s/system_cpu_percent.json' % self.instance_root: 
            '[  {    "entry": 10.0,     "time": "%s TIME"  }]' % date_to_test,
          '%s/system_net_out_bytes.json' % self.instance_root:
            '[  {    "entry": 0.0,     "time": "%s TIME"  }]' % date_to_test,
          '%s/system_net_in_bytes.json' % self.instance_root: 
            '[  {    "entry": 0.0,     "time": "%s TIME"  }]' % date_to_test,
          '%s/system_disk_memory_free__dev_sdx1.json' % self.instance_root:
            '[  {    "entry": 0.01953125,     "time": "%s TIME"  }, '\
             '  {    "entry": 0.01953125,     "time": "%s TIME"  }]' % (date_to_test, date_to_test),
          '%s/system_net_out_errors.json' % self.instance_root: 
            '[  {    "entry": 1.0,     "time": "%s TIME"  }]' % date_to_test,
          '%s/system_disk_memory_used__dev_sdx1.json' % self.instance_root: 
            '[  {    "entry": 0.009765625,     "time": "%s TIME"  },'\
            '   {    "entry": 0.009765625,     "time": "%s TIME"  }]' % (date_to_test, date_to_test),
          '%s/system_net_out_dropped.json' % self.instance_root: 
            '[  {    "entry": 1.0,     "time": "%s TIME"  }]' % date_to_test,
          '%s/system_memory_free.json' % self.instance_root:
            '[  {    "entry": 0.09765625,     "time": "%s TIME"  }]' % date_to_test,
          '%s/system_net_in_errors.json' % self.instance_root: 
            '[  {    "entry": 1.0,     "time": "%s TIME"  }]' % date_to_test,
          '%s/system_net_in_dropped.json' % self.instance_root: 
            '[  {    "entry": 2.0,     "time": "%s TIME"  }]' % date_to_test,
          '%s/system_loadavg.json' % self.instance_root: 
            '[  {    "entry": 0.1,     "time": "%s TIME"  }]' % date_to_test}

        self.assertEqual(set(glob.glob("%s/*.json" % self.instance_root)),
                         set(json_path_dict)) 

        for f_path in set(glob.glob("%s/*.json" % self.instance_root)):
          with open(f_path, "r") as value:
            self.assertEqual(json.load(value), json.loads(json_path_dict[f_path]))


    def test_system_csv_report(self):
        database = self.getPopulatedDB(datetime.utcnow().strftime("%Y-%m-%d"), amount=2)
        reporter.SystemCSVReporterDumper(database).dump(self.instance_root)
        date_for_test = datetime.utcnow().strftime("%Y-%m-%d")
        csv_path_dict = {'%s/system_memory_used.csv' % self.instance_root:
                           [['time','entry'], ['%s TIME' % date_for_test,'0.09765625']],
                         '%s/system_cpu_percent.csv' % self.instance_root:
                           [['time','entry'], ['%s TIME' % date_for_test,'10.0']],
                         '%s/system_net_out_bytes.csv' % self.instance_root:
                           [['time','entry'], ['%s TIME' % date_for_test,'0.0']],
                         '%s/system_net_in_bytes.csv' % self.instance_root:
                           [['time','entry'], ['%s TIME' % date_for_test,'0.0']],
                         '%s/system_disk_memory_free__dev_sdx1.csv' % self.instance_root: 
                           [['time','entry'], ['%s TIME' % date_for_test,'0.01953125'], 
                            ['%s TIME' % date_for_test,'0.01953125']],
                         '%s/system_net_out_errors.csv' % self.instance_root:
                           [['time','entry'], ['%s TIME' % date_for_test,'1.0']],
                         '%s/system_disk_memory_used__dev_sdx1.csv' % self.instance_root: 
                           [['time','entry'], ['%s TIME' % date_for_test,'0.009765625'], 
                             ['%s TIME' % date_for_test,'0.009765625']],
                         '%s/system_net_out_dropped.csv' % self.instance_root: 
                           [['time','entry'], ['%s TIME' % date_for_test,'1.0']],
                         '%s/system_memory_free.csv' % self.instance_root:
                           [['time','entry'], ['%s TIME' % date_for_test,'0.09765625']],
                         '%s/system_net_in_errors.csv' % self.instance_root:
                           [['time','entry'], ['%s TIME' % date_for_test,'1.0']],
                         '%s/system_net_in_dropped.csv' % self.instance_root: 
                           [['time','entry'], ['%s TIME' % date_for_test,'2.0']],
                         '%s/system_loadavg.csv' % self.instance_root:
                            [['time','entry'], ['%s TIME' % date_for_test,'0.1']]}

        self.assertEqual(set(glob.glob("%s/*.csv" % self.instance_root)),
                    set(csv_path_dict.keys())) 

        for f_path in list(set(glob.glob("%s/*.csv" % self.instance_root))):
          with open(f_path, "r") as csv_file:
            self.assertEqual([i for i in csv.reader(csv_file)], csv_path_dict[f_path])


    def test_compress_log_directory(self):
        log_directory = "%s/test_compress" % self.instance_root
        dump_folder = "%s/1990-01-01" % log_directory

        if not os.path.exists(log_directory):
            os.mkdir(log_directory)

        if os.path.exists(dump_folder):
            shutil.rmtree(dump_folder)

        os.mkdir("%s/1990-01-01" % log_directory)
        with open("%s/test.txt" % dump_folder, "w") as dump_file:
            dump_file.write("hi")
            dump_file.close()

        reporter.compressLogFolder(log_directory)

        self.assertFalse(os.path.exists(dump_folder))

        self.assertTrue(os.path.exists("%s.tar.gz" % dump_folder))

        with tarfile.open("%s.tar.gz" % dump_folder) as tf:
            self.assertEqual(tf.getmembers()[0].name, "1990-01-01")
            self.assertEqual(tf.getmembers()[1].name, "1990-01-01/test.txt")
            self.assertEqual(tf.extractfile(tf.getmembers()[1]).read(), b'hi')




class TestCollectSnapshot(unittest.TestCase):

    def setUp(self):
        self.slap = slapos.slap.slap()
        self.app = SlapOSApp()
        self.temp_dir = tempfile.mkdtemp()
        self.addCleanup(os.environ.__setitem__, "HOME", os.environ["HOME"])
        os.environ["HOME"] = self.temp_dir
        self.instance_root = tempfile.mkdtemp()
        self.software_root = tempfile.mkdtemp()
        if os.path.exists(self.temp_dir):
          shutil.rmtree(self.temp_dir)

    def getFakeUser(self, disk_snapshot_params={}):
       os.mkdir("%s/fakeuser0" % self.instance_root)
       return entity.User("fakeuser0", 
                    "%s/fakeuser0" % self.instance_root, disk_snapshot_params ) 

    def test_process_snapshot(self):
        process = psutil.Process(os.getpid())
        process_snapshot = snapshot.ProcessSnapshot(process)

        self.assertNotEqual(process_snapshot.username, None)  
        self.assertEqual(int(process_snapshot.pid), os.getpid())
        self.assertEqual(int(process_snapshot.process.split("-")[0]),
                          os.getpid())

        self.assertNotEqual(process_snapshot.cpu_percent , None)
        self.assertNotEqual(process_snapshot.cpu_time , None)
        self.assertNotEqual(process_snapshot.cpu_num_threads, None)
        self.assertNotEqual(process_snapshot.memory_percent , None)
        self.assertNotEqual(process_snapshot.memory_rss, None)
        self.assertNotEqual(process_snapshot.io_rw_counter, None)
        self.assertNotEqual(process_snapshot.io_cycles_counter, None)

    def test_folder_size_snapshot(self):
        disk_snapshot = snapshot.FolderSizeSnapshot(self.instance_root)
        self.assertEqual(disk_snapshot.disk_usage, 0)
        for i in range(0, 10):
          folder = 'folder%s' % i
          os.mkdir(os.path.join(self.instance_root, folder))
          with open(os.path.join(self.instance_root, folder, 'toto'), 'w') as f:
            f.write('toto text')

        disk_snapshot.update_folder_size()
        self.assertNotEqual(disk_snapshot.disk_usage, 0)

        pid_file = os.path.join(self.instance_root, 'disk_snap.pid')
        disk_snapshot = snapshot.FolderSizeSnapshot(self.instance_root, pid_file)
        disk_snapshot.update_folder_size()
        self.assertNotEqual(disk_snapshot.disk_usage, 0)

        pid_file = os.path.join(self.instance_root, 'disk_snap.pid')
        disk_snapshot = snapshot.FolderSizeSnapshot(self.instance_root, pid_file,
                                           use_quota=True)
        disk_snapshot.update_folder_size()
        self.assertNotEqual(disk_snapshot.disk_usage, 0)

    def test_process_in_progress_disk_usage(self):
        pid_file = os.path.join(self.instance_root, 'sleep.pid')
        disk_snapshot = snapshot.FolderSizeSnapshot(self.instance_root, pid_file)

        command = 'sleep 1h'
        process = subprocess.Popen(command, stdout=subprocess.PIPE,
                                  stderr=subprocess.PIPE, shell=True)
        
        with open(pid_file, 'w') as fpid:
          pid = fpid.write(str(process.pid))
        self.assertTrue(os.path.isfile(pid_file))

        self.assertEqual(disk_snapshot.update_folder_size(), None)

        disk_snapshot = snapshot.FolderSizeSnapshot(self.instance_root, pid_file,
                                           use_quota=True)
        self.assertEqual(disk_snapshot.update_folder_size(), None)

        process.terminate()

    def test_time_cycle(self):
        disk_snapshot_params = {'enable': True, 'time_cycle': 3600, 'testing': True}
        user = self.getFakeUser(disk_snapshot_params)
        database = db.Database(self.instance_root, create=True)

        date = datetime.utcnow().date()
        time = datetime.utcnow().time().strftime("%H:%M:%S")
        datetime_earlier = datetime.utcnow() - timedelta(hours=3)
        date_earlier = datetime_earlier.date()
        time_earlier = datetime_earlier.time().strftime("%H:%M:%S")

        database.connect()
        database.inserFolderSnapshot('fakeuser0', '1.0', date_earlier, time_earlier)
        database.commit()
        database.close()

        # check that _insertDiskSnapShot called update_folder_size
        with mock.patch('slapos.collect.snapshot.FolderSizeSnapshot.update_folder_size'
                        ) as update_folder_size_call:
          user._insertDiskSnapShot(database, date, time)
          update_folder_size_call.assert_called_once()

        datetime_earlier = datetime.utcnow() - timedelta(minutes=10)
        date_earlier = datetime_earlier.date()
        time_earlier = datetime_earlier.time().strftime("%H:%M:%S")
        
        database.connect()
        database.inserFolderSnapshot('fakeuser0', '1.0', date_earlier, time_earlier)
        database.commit()
        database.close()

        # check that _insertDiskSnapShot stop before calling update_folder_size
        with mock.patch('slapos.collect.snapshot.FolderSizeSnapshot.update_folder_size'
                        ) as update_folder_size_call:
          user._insertDiskSnapShot(database, date, time)
          update_folder_size_call.assert_not_called()


    def test_process_snapshot_broken_process(self):
        self.assertRaises(AssertionError, 
                 snapshot.ProcessSnapshot, None)

    def test_computer_snapshot(self):
        computer_snapshot = snapshot.ComputerSnapshot()
        self.assertNotEqual(computer_snapshot.cpu_num_core , None)
        self.assertNotEqual(computer_snapshot.cpu_frequency , None)
        self.assertNotEqual(computer_snapshot.cpu_type , None)
        self.assertNotEqual(computer_snapshot.memory_size , None)
        self.assertNotEqual(computer_snapshot.memory_type , None)
        
        self.assertEqual(type(computer_snapshot.system_snapshot),  
                               snapshot.SystemSnapshot)

        self.assertNotEqual(computer_snapshot.disk_snapshot_list, [])
        self.assertNotEqual(computer_snapshot.partition_list, []) 

        self.assertEqual(type(computer_snapshot.disk_snapshot_list[0]), 
                snapshot.DiskPartitionSnapshot)

    def test_system_snapshot(self):
        system_snapshot = snapshot.SystemSnapshot()       
        self.assertNotEqual(system_snapshot.memory_used , None)
        self.assertNotEqual(system_snapshot.memory_free , None)
        self.assertNotEqual(system_snapshot.memory_percent , None)
        self.assertNotEqual(system_snapshot.cpu_percent , None)
        self.assertNotEqual(system_snapshot.load , None)
        self.assertNotEqual(system_snapshot.net_in_bytes , None)
        self.assertNotEqual(system_snapshot.net_in_errors, None)
        self.assertNotEqual(system_snapshot.net_in_dropped , None)
        self.assertNotEqual(system_snapshot.net_out_bytes , None)
        self.assertNotEqual(system_snapshot.net_out_errors, None)
        self.assertNotEqual(system_snapshot.net_out_dropped , None)
 
class TestCollectEntity(unittest.TestCase):

    def setUp(self):
        self.temp_dir = tempfile.mkdtemp()
        self.addCleanup(os.environ.__setitem__, "HOME", os.environ["HOME"])
        os.environ["HOME"] = self.temp_dir
        self.instance_root = tempfile.mkdtemp()
        self.software_root = tempfile.mkdtemp()
        if os.path.exists(self.temp_dir):
          shutil.rmtree(self.temp_dir)

    def tearDown(self):
        pass

    def getFakeUser(self, disk_snapshot_params={}):
       os.mkdir("%s/fakeuser0" % self.instance_root)
       return entity.User("fakeuser0", 
                    "%s/fakeuser0" % self.instance_root, disk_snapshot_params ) 

    def test_get_user_list(self):
        config = ConfigParser()
        config.add_section('slapformat')
        config.set('slapformat', 'partition_amount', '3')
        config.set('slapformat', 'user_base_name', 'slapuser')
        config.set('slapformat', 'partition_base_name', 'slappart')
        config.add_section('slapos')
        config.set('slapos', 'instance_root', self.instance_root)
 
        user_dict = entity.get_user_list(config)
        username_set = {'slapuser0', 'slapuser1', 'slapuser2'} 
        self.assertEqual(username_set, set(user_dict))
       
        for name in username_set:
          self.assertEqual(user_dict[name].name, name)
          self.assertEqual(user_dict[name].snapshot_list, [])
          expected_path = "%s/slappart%s" % (self.instance_root, name.strip("slapuser")) 
          self.assertEqual(user_dict[name].path, expected_path) 
       
    def test_user_add_snapshot(self):
        user = self.getFakeUser() 
        self.assertEqual(user.snapshot_list, [])
        user.append("SNAPSHOT")
        self.assertEqual(user.snapshot_list, ["SNAPSHOT"])

    def test_user_save(self):
        disk_snapshot_params = {'enable': False}
        user = self.getFakeUser(disk_snapshot_params)
        process = psutil.Process(os.getpid())
        user.append(snapshot.ProcessSnapshot(process))
        database = FakeDatabase()
        user.save(database, "DATE", "TIME")
        self.assertEqual(database.invoked_method_list[0], ("connect", ""))

        self.assertEqual(database.invoked_method_list[1][0], "insertUserSnapshot")
        self.assertEqual(database.invoked_method_list[1][1][0], ("fakeuser0",))
        self.assertEqual(set(database.invoked_method_list[1][1][1]), 
                   {'cpu_time', 'cpu_percent', 'process',
                    'memory_rss', 'pid', 'memory_percent',
                    'io_rw_counter', 'insertion_date', 'insertion_time',
                    'io_cycles_counter', 'cpu_num_threads'})
        self.assertEqual(database.invoked_method_list[2], ("commit", ""))
        self.assertEqual(database.invoked_method_list[3], ("close", ""))

    def test_user_save_disk_snapshot(self):
        disk_snapshot_params = {'enable': True, 'testing': True}
        user = self.getFakeUser(disk_snapshot_params)
        process = psutil.Process(os.getpid())
        user.append(snapshot.ProcessSnapshot(process))
        database = FakeDatabase2()
        user.save(database, "DATE", "TIME")
        self.assertEqual(database.invoked_method_list[0], ("connect", ""))

        self.assertEqual(database.invoked_method_list[1][0], "insertUserSnapshot")
        self.assertEqual(database.invoked_method_list[1][1][0], ("fakeuser0",))
        self.assertEqual(set(database.invoked_method_list[1][1][1]), 
                   {'cpu_time', 'cpu_percent', 'process',
                    'memory_rss', 'pid', 'memory_percent',
                    'io_rw_counter', 'insertion_date', 'insertion_time',
                    'io_cycles_counter', 'cpu_num_threads'})
        self.assertEqual(database.invoked_method_list[2], ("commit", ""))
        self.assertEqual(database.invoked_method_list[3], ("close", ""))

        self.assertEqual(database.invoked_method_list[4], ("connect", ""))
        self.assertEqual(database.invoked_method_list[5][0], "inserFolderSnapshot")
        self.assertEqual(database.invoked_method_list[5][1][0], ("fakeuser0",))
        self.assertEqual(set(database.invoked_method_list[5][1][1]), 
                   {'insertion_date', 'disk_usage', 'insertion_time'})
        self.assertEqual(database.invoked_method_list[6], ("commit", ""))
        self.assertEqual(database.invoked_method_list[7], ("close", ""))

    def test_user_save_disk_snapshot_cycle(self):
        disk_snapshot_params = {'enable': True, 'time_cycle': 3600, 'testing': True}
        user = self.getFakeUser(disk_snapshot_params)
        process = psutil.Process(os.getpid())
        user.append(snapshot.ProcessSnapshot(process))
        database = FakeDatabase2()
        user.save(database, "DATE", "TIME")
        self.assertEqual(database.invoked_method_list[0], ("connect", ""))

        self.assertEqual(database.invoked_method_list[1][0], "insertUserSnapshot")
        self.assertEqual(database.invoked_method_list[1][1][0], ("fakeuser0",))
        self.assertEqual(set(database.invoked_method_list[1][1][1]), 
                   {'cpu_time', 'cpu_percent', 'process',
                    'memory_rss', 'pid', 'memory_percent',
                    'io_rw_counter', 'insertion_date', 'insertion_time',
                    'io_cycles_counter', 'cpu_num_threads'})
        self.assertEqual(database.invoked_method_list[2], ("commit", ""))
        self.assertEqual(database.invoked_method_list[3], ("close", ""))

        self.assertEqual(database.invoked_method_list[4], ("connect", ""))
        self.assertEqual(database.invoked_method_list[5][0], "select")
        self.assertEqual(database.invoked_method_list[5][1][0], ())
        self.assertEqual(set(database.invoked_method_list[5][1][1]),
                                {'table', 'where', 'limit', 'order', 'columns'})
        self.assertEqual(database.invoked_method_list[6][0], "inserFolderSnapshot")
        self.assertEqual(database.invoked_method_list[6][1][0], ("fakeuser0",))
        self.assertEqual(set(database.invoked_method_list[6][1][1]), 
                   {'insertion_date', 'disk_usage', 'insertion_time'})
        self.assertEqual(database.invoked_method_list[7], ("commit", ""))
        self.assertEqual(database.invoked_method_list[8], ("close", ""))

    def test_computer_entity(self):
        computer = entity.Computer(snapshot.ComputerSnapshot())
        database = FakeDatabase()
        computer.save(database, "DATE", "TIME")

        self.assertEqual(database.invoked_method_list[0], ("connect", ""))

        self.assertEqual(database.invoked_method_list[1][0], "insertComputerSnapshot")
        self.assertEqual(database.invoked_method_list[1][1][0], ())
        self.assertEqual(set(database.invoked_method_list[1][1][1]), 
                 {'insertion_time', 'insertion_date', 'cpu_num_core',
                  'partition_list', 'cpu_frequency', 'memory_size', 
                  'cpu_type', 'memory_type'})
 
        self.assertEqual(database.invoked_method_list[2][0], "insertSystemSnapshot")
        self.assertEqual(database.invoked_method_list[2][1][0], ())
        self.assertEqual(set(database.invoked_method_list[2][1][1]), 
          set([ 'memory_used', 'cpu_percent', 'insertion_date', 'insertion_time',
                'loadavg', 'memory_free', 'net_in_bytes', 'net_in_dropped', 
                'net_in_errors', 'net_out_bytes', 'net_out_dropped', 
                'net_out_errors']))

        self.assertEqual(database.invoked_method_list[3][0], "insertDiskPartitionSnapshot")
        self.assertEqual(database.invoked_method_list[3][1][0], ())
        self.assertEqual(set(database.invoked_method_list[3][1][1]), 
          set([ 'used', 'insertion_date', 'partition', 'free', 
                'mountpoint', 'insertion_time' ]))

        self.assertEqual(database.invoked_method_list[-2], ("commit", ""))
        self.assertEqual(database.invoked_method_list[-1], ("close", ""))

class TestJournal(unittest.TestCase):

  def test_journal(self):
    journal = reporter.Journal()

    self.assertEqual(journal.getXML(),
      b"<?xml version='1.0' encoding='utf-8'?><journal/>")

    transaction = journal.newTransaction()
    journal.setProperty(transaction, "title", "TestJournal")
    journal.setProperty(transaction, "reference", "reference-of-testjournal")

    arrow = ElementTree.SubElement(transaction, "arrow")
    arrow.set("type", "Destination")

    journal.newMovement(transaction,
                        resource="ee",
                        title="ZZ",
                        quantity="10",
                        reference="BB",
                        category="")


    self.assertEqual(journal.getXML(),
      b'<?xml version=\'1.0\' encoding=\'utf-8\'?><journal><transaction type="Sale Packing List"><title>TestJournal</title><reference>reference-of-testjournal</reference><arrow type="Destination"/><movement><resource>ee</resource><title>ZZ</title><reference>BB</reference><quantity>10</quantity><price>0.0</price><VAT></VAT><category></category></movement></transaction></journal>')


class TestConsumptionReportBase(unittest.TestCase):

  base_path, = slapos.tests.data.__path__
  maxDiff = None

  def _get_file_content(self, f_path):
    with open(f_path, "r") as f:
      return f.readlines()

  def loadPredefinedDB(self):
    # populate db
    conn = sqlite3.connect(
      os.path.join(self.instance_root, 'collector.db'))
    with open(os.path.join(self.base_path, "monitor_collect.sql")) as f:
      sql = f.read()
    conn.executescript(sql)
    conn.close() 

  def get_fake_user_list(self, partition_amount=3):
    config = ConfigParser()
    config.add_section('slapformat')
    config.set('slapformat', 'partition_amount', str(partition_amount))
    config.set('slapformat', 'user_base_name', 'slapuser')
    config.set('slapformat', 'partition_base_name', 'slappart')
    config.add_section('slapos')
    config.set('slapos', 'instance_root', self.instance_root)
    
    return entity.get_user_list(config)

  def _getReport(self):
    return reporter.ConsumptionReportBase(self.database)

  def setUp(self):
    self.instance_root = tempfile.mkdtemp()
    # inititalise
    self.loadPredefinedDB()
    self.database = db.Database(self.instance_root, create=True)
    self.temp_dir = tempfile.mkdtemp()
    self.addCleanup(os.environ.__setitem__, "HOME", os.environ["HOME"])
    os.environ["HOME"] = self.temp_dir
    self.software_root = tempfile.mkdtemp()

    self.report = self._getReport() 

  def test_getPartitionUsedMemoryAverage(self):
    self.assertEqual(None,
      self.report.getPartitionUsedMemoryAverage('slapuser19', '2019-10-04'))
    self.assertEqual(None,
      self.report.getPartitionUsedMemoryAverage('slapuser15', '2019-10-05'))
    self.assertAlmostEqual(3868924121.87234,
      self.report.getPartitionUsedMemoryAverage('slapuser19', '2019-10-05'))

  def test_getPartitionCPULoadAverage(self):
    self.assertAlmostEqual(7.08297872340419,
      self.report.getPartitionCPULoadAverage('slapuser19', '2019-10-05'))
    self.assertEqual(None,
      self.report.getPartitionCPULoadAverage('slapuser15', '2019-10-05'))
    self.assertEqual(None,
      self.report.getPartitionCPULoadAverage('slapuser19', '2019-10-04'))


  def test_getPartitionDiskUsedAverage(self):
    self.assertAlmostEqual(7693240.0,
      self.report.getPartitionDiskUsedAverage('slapuser19', '2019-10-05'))
    self.assertEqual(None,
      self.report.getPartitionDiskUsedAverage('slapuser99', '2019-10-05'))

  def test_getPartitionProcessConsumptionList(self):
    data = self.report.getPartitionProcessConsumptionList(
            'slapuser19', date_scope='2019-10-05', 
            min_time='00:01:00', max_time='00:13:00')
    self.assertAlmostEqual(0.02, data[-1]['cpu_time'])
    self.assertAlmostEqual(1363.0, data[-1]['pid'])
    self.assertAlmostEqual(193206.0, data[-1]['io_cycles_counter'])
    self.assertAlmostEqual(5.16, data[-1]['memory_rss'])
    self.assertAlmostEqual(2916352.0, data[-1]['io_rw_counter'])

  def test_getPartitionConsumptionStatusList(self):
    data = self.report.getPartitionConsumptionStatusList('slapuser19', date_scope='2019-10-05',
                                         min_time='00:01:00', max_time='00:13:00')
    self.assertAlmostEqual(14.6, data[0]['cpu_percent'])
    self.assertAlmostEqual(2773721862147.0, data[2]['io_rw_counter'])

class TestConsumptionReport(TestConsumptionReportBase):

  def _getReport(self):
    return reporter.ConsumptionReport(database=self.database,
                                      computer_id="COMP-192938",
                                      location=self.temp_dir,
                                      user_list=self.get_fake_user_list(15))

  def test_getCpuLoadAverageConsumption(self):
    self.assertAlmostEqual(self.report._getCpuLoadAverageConsumption('2019-10-05'), 74.44468085106385)
    self.assertAlmostEqual(self.report._getCpuLoadAverageConsumption('2019-10-06'), 74.99183673469388)
    self.assertAlmostEqual(self.report._getCpuLoadAverageConsumption('2019-10-07'), 76.43714285714287)
    self.assertEqual(self.report._getCpuLoadAverageConsumption('2019-10-08'), None)
    self.assertEqual(self.report._getCpuLoadAverageConsumption('2019-NO(d'), None)

  def test_getMemoryAverageConsumption(self):
    self.assertAlmostEqual(self.report._getMemoryAverageConsumption('2019-10-05'), 14185032159.319149)
    self.assertAlmostEqual(self.report._getMemoryAverageConsumption('2019-10-06'), 14149247895.510204)
    self.assertAlmostEqual(self.report._getMemoryAverageConsumption('2019-10-07'), 14211174517.028572)
    self.assertEqual(self.report._getMemoryAverageConsumption('2019-10-08'), None)
    self.assertEqual(self.report._getMemoryAverageConsumption('2019-NO(d'), None)


  def test_buildXMLReport(self):
    with open(os.path.join(self.temp_dir, "2019-10-07.xml.uploaded"), "w+") as f:
      f.write("")
    report_path = self.report.buildXMLReport('2019-10-07')
    self.assertEqual(report_path, None)
    os.unlink(os.path.join(self.temp_dir, "2019-10-07.xml.uploaded"))

    report_path = self.report.buildXMLReport('2019-10-07')
    self.assertEqual(report_path, os.path.join(self.temp_dir, "2019-10-07.xml"))
    expected_file_path = "%s/%s" % (self.base_path, "2019-10-07.xml")
    with open(report_path, "r") as report_file:
      with open(expected_file_path, "r") as expected_file: 
        self.assertEqual(slapos.util.xml2dict(report_file.read()),
                         slapos.util.xml2dict(expected_file.read()))



class TestPartitionReport(TestConsumptionReportBase):
  def _getReport(self):
    return reporter.PartitionReport(self.database,
                                    user_list=self.get_fake_user_list(20))

  def test_initDataFile(self):
    now = time.time()
    time.sleep(.1)
    self.report._initDataFile(
      os.path.join(self.temp_dir, "testfile.json"), ["my", "colums", "goes", "here"])
    time.sleep(.1)
    with open(os.path.join(self.temp_dir, "testfile.json")) as f:
      _json = json.load(f)

    self.assertLessEqual(_json["date"], time.time())
    self.assertLessEqual(now , _json["date"])
    self.assertEqual(["my", "colums", "goes", "here"], _json["data"])
    self.assertEqual(["data", "date"], sorted(_json))

  def test_buildJSONMonitorReport(self):
    with mock.patch('time.time', return_value=1570495649.5), mock.patch('psutil.Process', side_effect=psutil.NoSuchProcess(1)):
      self.report.buildJSONMonitorReport()
      
      for user in self.report.user_list.values():
        location = os.path.join(user.path, ".slapgrid")
        self.assertFalse(set(glob.glob("%s/*.json" % location))) 
  
      for user in self.report.user_list.values():
        location = os.path.join(user.path, ".slapgrid")
        if not os.path.exists(location):
          if user.name in ["slapuser19", "slapuser12"]: 
            mkdir_p(location, 0o755)
        else:
          if user.name not in ["slapuser19", "slapuser12"]: 
            shutil.rmtree(location)
      self.report.buildJSONMonitorReport(date_scope="2019-10-06",
                                         min_time="00:00:00",
                                         max_time="00:10:00")

      for user in self.report.user_list.values():
        if user.name not in ["slapuser19", "slapuser12"]:
          continue
        csv_path_list = [
          '%s/.slapgrid/monitor/monitor_resource.status.json' % user.path, 
          '%s/.slapgrid/monitor/monitor_resource_memory.data.json' % user.path,
          '%s/.slapgrid/monitor/monitor_process_resource.status.json' % user.path,
          '%s/.slapgrid/monitor/monitor_resource_process.data.json' % user.path,
          '%s/.slapgrid/monitor/monitor_resource_io.data.json' % user.path] 
  
        self.assertEqual(set(glob.glob("%s/.slapgrid/monitor/*.json" % user.path)),
                         set(csv_path_list)) 
     
        for f_path in set(glob.glob("%s/.slapgrid/monitor/*.json" % user.path)):
          expected_file_path = os.path.join(self.base_path, user.name, os.path.basename(f_path))
          #if six.PY2:
          #  mkdir_p("%s/%s" % (self.base_path, user.name))
          #  with open(f_path) as f:
          #    with open(expected_file_path, "w") as z:
          #      z.write(f.read())
          with open(f_path, "r") as value:
            with open(expected_file_path, "r") as expected:
              json_value = json.load(value)
              json_expected = json.load(expected)
              self.assertEqual(json_value, json_expected) 



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2024 Nexedi and Contributors. All Rights Reserved.
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsibility of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# guarantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

"""Utility classes to use caucase and certificates in tests."""

import hashlib
import os
import shutil
import subprocess
import tempfile
import time
from typing import Optional
import urllib.parse

from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.x509.oid import NameOID
import requests

from .utils import findFreeTCPPortRange, ManagedResource


class CaucaseService(ManagedResource):
  """A caucase service."""

  url: str
  directory: str
  _caucased_process: subprocess.Popen

  @property
  def ca_crt_path(self) -> str:
    """Path of the CA certificate from this caucase."""
    ca_crt_path = os.path.join(self.directory, "ca.crt.pem")
    if not os.path.exists(ca_crt_path):
      with open(ca_crt_path, "w") as f:
        f.write(
          requests.get(
            urllib.parse.urljoin(
              self.url,
              "/cas/crt/ca.crt.pem",
            )
          ).text
        )
    return ca_crt_path

  @property
  def _caucased_path(self) -> str:
    """path of caucased executable.

    Expects the software release to have `bin/caucased`
    """
    software_release_root_path = os.path.join(
      self._cls.slap._software_root,
      hashlib.md5(self._cls.getSoftwareURL().encode()).hexdigest(),
    )
    return os.path.join(software_release_root_path, "bin", "caucased")

  def open(self) -> None:
    # starts a caucased.
    self.directory = tempfile.mkdtemp()
    caucased_dir = os.path.join(self.directory, "caucased")
    os.mkdir(caucased_dir)
    os.mkdir(os.path.join(caucased_dir, "user"))
    os.mkdir(os.path.join(caucased_dir, "service"))

    backend_caucased_netloc = f"{self._cls._ipv4_address}:{findFreeTCPPortRange(self._cls._ipv4_address, 2)}"
    self.url = f"http://{backend_caucased_netloc}"
    self._caucased_process = subprocess.Popen(
      [
        self._caucased_path,
        "--db",
        os.path.join(caucased_dir, "caucase.sqlite"),
        "--server-key",
        os.path.join(caucased_dir, "server.key.pem"),
        "--netloc",
        backend_caucased_netloc,
        "--service-auto-approve-count",
        "1",
      ],
      # capture subprocess output not to pollute test's own stdout
      stdout=subprocess.PIPE,
      stderr=subprocess.STDOUT,
    )
    for _ in range(30):
      try:
        if requests.get(self.url).ok:
          break
      except Exception:
        pass
      time.sleep(1)
    else:
      raise RuntimeError("caucased failed to start.")

  def close(self) -> None:
    self._caucased_process.terminate()
    self._caucased_process.wait()
    assert self._caucased_process.stdout
    self._caucased_process.stdout.close()
    shutil.rmtree(self.directory)


class CaucaseCertificate(ManagedResource):
  """A certificate signed by a caucase service."""

  ca_crt_file: str
  crl_file: str
  csr_file: str
  cert_file: str
  key_file: str

  def open(self) -> None:
    self.tmpdir = tempfile.mkdtemp()
    self.ca_crt_file = os.path.join(self.tmpdir, "ca-crt.pem")
    self.crl_file = os.path.join(self.tmpdir, "ca-crl.pem")
    self.csr_file = os.path.join(self.tmpdir, "csr.pem")
    self.cert_file = os.path.join(self.tmpdir, "crt.pem")
    self.key_file = os.path.join(self.tmpdir, "key.pem")

  def close(self) -> None:
    shutil.rmtree(self.tmpdir)

  @property
  def _caucase_path(self) -> str:
    """path of caucase executable."""
    software_release_root_path = os.path.join(
      self._cls.slap._software_root,
      hashlib.md5(self._cls.getSoftwareURL().encode()).hexdigest(),
    )
    return os.path.join(software_release_root_path, "bin", "caucase")

  def request(
    self,
    common_name: str,
    caucase: CaucaseService,
    san: Optional[x509.SubjectAlternativeName] = None,
  ) -> None:
    """Generate certificate and request signature to the caucase service.

    This overwrite any previously requested certificate for this instance.
    """
    cas_args = [
      self._caucase_path,
      "--ca-url",
      caucase.url,
      "--ca-crt",
      self.ca_crt_file,
      "--crl",
      self.crl_file,
    ]

    key = rsa.generate_private_key(
      public_exponent=65537, key_size=2048, backend=default_backend()
    )
    with open(self.key_file, "wb") as f:
      f.write(
        key.private_bytes(  # type:ignore
          encoding=serialization.Encoding.PEM,
          format=serialization.PrivateFormat.TraditionalOpenSSL,
          encryption_algorithm=serialization.NoEncryption(),
        )
      )

    csr = x509.CertificateSigningRequestBuilder().subject_name(
      x509.Name(
        [
          x509.NameAttribute(
            NameOID.COMMON_NAME,
            common_name,
          ),
        ]
      )
    )
    if san:
      csr = csr.add_extension(san, critical=True)
    csr = csr.sign(key, hashes.SHA256(), default_backend())
    with open(self.csr_file, "wb") as f:
      f.write(csr.public_bytes(serialization.Encoding.PEM))

    csr_id = (
      subprocess.check_output(
        cas_args
        + [
          "--send-csr",
          self.csr_file,
        ],
      )
      .split()[0]
      .decode()
    )
    assert csr_id

    for _ in range(30):
      if (
        not subprocess.call(
          cas_args
          + [
            "--get-crt",
            csr_id,
            self.cert_file,
          ],
          stdout=subprocess.PIPE,
          stderr=subprocess.STDOUT,
        )
        == 0
      ):
        break
      else:
        time.sleep(1)
    else:
      raise RuntimeError("getting service certificate failed.")
    with open(self.cert_file) as cert_file:
      assert "BEGIN CERTIFICATE" in cert_file.read()

  def revoke(self, caucase: CaucaseService) -> None:
    """Revoke the client certificate on this caucase instance."""
    subprocess.check_call(
      [
        self._caucase_path,
        "--ca-url",
        caucase.url,
        "--ca-crt",
        self.ca_crt_file,
        "--crl",
        self.crl_file,
        "--revoke-crt",
        self.cert_file,
        self.key_file,
      ]
    )



# -*- coding: utf-8 -*-
# vim: set et sts=2:
##############################################################################
#
# Copyright (c) 2010, 2011, 2012, 2013, 2014 Vifib SARL and Contributors.
# All Rights Reserved.
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsibility of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# guarantees and support are strongly advised to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

import re
import os
from datetime import datetime
from slapos.proxy.db_version import DB_VERSION
from slapos.util import sqlite_connect

from flask import g, Flask, redirect, url_for, current_app
from werkzeug.middleware.proxy_fix import ProxyFix
from .hateoas import hateoas_blueprint
from .slap_tool import slap_tool_blueprint
from .http_proxy import http_proxy_blueprint
from .db import execute_db
from .json_rpc import JsonRpcManager
from .panel import panel_blueprint

from six.moves.urllib.parse import urlparse

app = Flask(__name__)
# Support having haproxy/nginx in front to provide https (in this case haproxy should set X-Forwarded-Proto header)
# Support to be in a different directory than / (in this case haproxy should set X-Forwarded-Prefix header)
app.wsgi_app = ProxyFix(app.wsgi_app, x_proto=1, x_prefix=1)
app.register_blueprint(hateoas_blueprint, url_prefix="/hateoas")
app.register_blueprint(slap_tool_blueprint)
app.register_blueprint(http_proxy_blueprint, url_prefix="/http_proxy")
JsonRpcManager().init_app(app)
app.register_blueprint(panel_blueprint, url_prefix="/panel")

def connect_db():
  return sqlite_connect(current_app.config['DATABASE_URI'])

def _upgradeDatabaseIfNeeded():
  """
  Analyses current database compared to defined schema,
  and adapt tables/data it if needed.
  """
  previous_table_list = g.db.execute(
    "SELECT name FROM sqlite_master WHERE type='table' ORDER BY Name"
  ).fetchall()
  if previous_table_list:
    search = re.compile(r'\d+$').search
    current_schema_version, = {search(table).group(0)
                               for table, in previous_table_list}
    # If version of current database is not old, do nothing
    if current_schema_version == DB_VERSION:
      return

    # first, make a backup of current database
    backup_file_name = "{}-backup-{}to{}-{}.sql".format(
        current_app.config['DATABASE_URI'],
        current_schema_version,
        DB_VERSION,
        datetime.now().isoformat())
    current_app.logger.info(
        'Old schema detected: Creating a backup of current tables at %s',
        backup_file_name
    )
    with open(backup_file_name, 'w') as f:
      for line in g.db.iterdump():
          f.write('%s\n' % line)

  with current_app.open_resource('schema.sql', 'r') as f:
    schema = f.read() % dict(version=DB_VERSION, computer=current_app.config['computer_id'])
  g.db.execute('BEGIN')
  try:
    g.db.executescript(schema)

    if previous_table_list:
      current_app.logger.info('Old schema detected: Migrating old tables...')
      n = len(current_schema_version)
      current_schema_version = int(current_schema_version)
      for old_table, in previous_table_list:
        rv = execute_db(old_table, 'SELECT * from %s', db_version='')
        if rv:
          table = old_table[:-n]
          if current_schema_version < 17:
            if table == 'local_software_release_root':
              path, = {row['path'] for row in rv}
              rv = {'name': table, 'value': path},
              table = 'config'
            elif table == 'partition':
              request_dict = {row['reference']: (i, row['requested_by'])
                              for i, row in enumerate(rv)}
              for i, requested_by in request_dict.values():
                if requested_by:
                  while True:
                    j, requested_by = request_dict[requested_by]
                    if not requested_by:
                      break
                  rv[i]['requested_by'] = rv[j]['partition_reference']
          for row in rv:
            query = 'INSERT OR REPLACE INTO %%s (%s) VALUES (:%s)' % (
              ', '.join(row), ', :'.join(row))
            execute_db(table, query, row)
        g.db.execute("DROP table " + old_table)
  except:
    g.db.rollback()
    raise
  g.db.commit()

def _updateLocalSoftwareReleaseRootPathIfNeeded():
  """
  Update the local software release root path if it changed,
  and rebase all URLs in the database relatively to the new path.
  """
  # Retrieve the current root path and replace it with the new one
  current_root_path = (execute_db('config',
    "SELECT value FROM %s WHERE name='local_software_release_root'",
    one=True) or {}).get('value', os.sep)
  new_root_path = current_app.config.get('local_software_release_root', os.sep)
  # Check whether one is the same as or a subpath of the other
  if current_root_path == new_root_path:
    return
  execute_db('config',
    "INSERT OR REPLACE INTO %s VALUES('local_software_release_root',?)",
    [new_root_path])
  relpath = os.path.relpath(new_root_path, current_root_path)
  if not relpath.startswith(os.pardir + os.sep):
    current_app.logger.info('Do not rebase any URLs because %s is a subpath of %s', new_root_path, current_root_path)
    return
  elif os.path.basename(relpath) == os.pardir:
    current_app.logger.info('Do not rebase any URLs because %s is a superpath of %s', new_root_path, current_root_path)
    return
  # Backup the database before migrating
  database_path = current_app.config['DATABASE_URI']
  backup_path = database_path + "-backup-%s.sql" % datetime.now().isoformat()
  current_app.logger.info("Backuping database to %s", backup_path)
  with open(backup_path, 'w') as f:
    for line in g.db.iterdump():
      f.write('%s\n' % line)
  # Rebase all URLs relative to the new root path
  current_app.logger.info('Rebase URLs on local software release root path')
  current_app.logger.info('Old root path: %s', current_root_path)
  current_app.logger.info('New root path: %s', new_root_path)
  def migrate_url(url):
    current_app.logger.debug('Examining URL %s', url)
    if not url or urlparse(url).scheme:
      current_app.logger.debug('  Do not rebase because it is not a path')
      return url
    rel = os.path.relpath(url, current_root_path)
    if rel.startswith(os.pardir + os.sep):
      current_app.logger.debug('  Do not rebase because it is not a subpath of %s', current_root_path)
      return url
    new = os.path.join(new_root_path, rel)
    if not os.path.isfile(new) and os.path.isfile(url):
      current_app.logger.debug('  Do not rebase because it refers to an existing file but %s does not', new)
      return url
    current_app.logger.debug('  Migrate to rebased URL %s', new)
    return new
  g.db.create_function('migrate_url', 1, migrate_url)
  execute_db('software', 'UPDATE %s SET url=migrate_url(url)')
  execute_db('partition', 'UPDATE %s SET software_release=migrate_url(software_release)')

is_schema_already_executed = False
@app.before_request
def before_request():
  g.db = connect_db()
  global is_schema_already_executed
  if not is_schema_already_executed:
    _upgradeDatabaseIfNeeded()
    _updateLocalSoftwareReleaseRootPathIfNeeded()
    is_schema_already_executed = True


@app.after_request
def after_request(response):
  if getattr(g, "db", None) is not None:
    # Only close the DB if it has been connected before
    g.db.commit()
    g.db.close()
  return response

@app.route('/getRunId', methods=['GET'])
def getRunId():
  return current_app.config['run_id']

@app.route('/', methods=['GET'])
def index():
  return redirect(url_for('panel.index'))



from slapos.tests.test_slapproxy import BasicMixin
import os
import unittest
import tempfile
import shutil


class TemporaryDirectory(object):
    """
    Context manager for tempfile.mkdtemp().
    This class is available in python +v3.2.
    """
    def __enter__(self):
        self.dir_name = tempfile.mkdtemp()
        return self.dir_name

    def __exit__(self, exc_type, exc_value, traceback):
        shutil.rmtree(self.dir_name)

TemporaryDirectory = getattr(tempfile, 'TemporaryDirectory',
                             TemporaryDirectory)


class PanelTestCase(BasicMixin, unittest.TestCase):

  def test_index_page(self):
    response = self.app.get('/panel/')

    # Response status code
    assert response.status_code == 200, response.status_code

    # Response body
    assert b'<script src="panel.js">' in response.data, response.data
    assert b'<script src="external/renderjs.js">' in response.data, response.data

    # Response header forced
    assert response.headers["Content-Type"] == \
        "text/html; charset=utf-8", \
        response.headers
    assert response.headers["Content-Length"] == \
        "3105", \
        response.headers

  def test_panel_dependencies(self):
    response = self.app.get('/panel/panel.js')

    # Response status code
    assert response.status_code == 200, response.status_code

    # Response body
    assert b'callJsonRpcEntryPoint' in response.data, response.data

    # Response header forced
    assert response.headers["Content-Type"] in \
        ["application/javascript; charset=utf-8", \
         "text/javascript; charset=utf-8"], \
        response.headers
    assert response.headers["Content-Length"] == \
        "32035", \
        response.headers

  def test_panel_external_dependencies(self):
    response = self.app.get('/panel/external/renderjs.js')

    # Response status code
    assert response.status_code == 200, response.status_code

    # Response body
    assert b'clearGadgetInternalParameters' in response.data, response.data

    # Response header forced
    assert response.headers["Content-Type"] in \
        ["application/javascript; charset=utf-8", \
         "text/javascript; charset=utf-8"], \
        response.headers
    assert response.headers["Content-Length"] == \
        "107174", \
        response.headers

  def test_panel_public_directory_not_configured(self):
    response = self.app.get('/panel/public/foo')
    assert response.status_code == 403, response.status_code
    assert b'The public directory path is not configured.' in response.data, response.data

  def test_panel_public_directory_configured(self):
    with TemporaryDirectory() as tmpdirname:
      with tempfile.NamedTemporaryFile(dir=tmpdirname) as fp:
        fp.write(b'Hello world!')
        fp.seek(0)
        base_name = os.path.basename(fp.name)
        self.app_config['PUBLIC_DIRECTORY_PATH'] = tmpdirname
        response = self.app.get('/panel/public/' + base_name)
    self.app_config.pop('PUBLIC_DIRECTORY_PATH')
    assert response.status_code == 200, response.status_code
    assert b'Hello world!' == response.data, response.data
    assert response.headers["Content-Disposition"] == "attachment; filename=%s" % base_name, response.headers



from werkzeug.routing import BaseConverter


class AbsolutePathConverter(BaseConverter):
    """Like the default :class:`UnicodeConverter`, but it also matches
    slashes. Requires to start with a slash
    """

    part_isolating = False
    regex = "/.*?"
    weight = 300



# -*- coding: utf-8 -*-
# vim: set et sts=2:
##############################################################################
#
# Copyright (c) 2010, 2011, 2012, 2013 Vifib SARL and Contributors.
# All Rights Reserved.
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsibility of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# guarantees and support are strongly advised to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

from slapos.proxy.views import app
from slapos.util import sqlite_connect

import random
import six
import string

def _generateSoftwareProductListFromString(software_product_list_string):
  """
  Take a string as argument (which usually comes from the software_product_list
  parameter of the slapproxy configuration file), and parse it to generate
  list of Software Products that slapproxy will use.
  """
  try:
    software_product_string_split = software_product_list_string.split('\n')
  except AttributeError:
    return {}
  software_product_list = {}
  for line in software_product_string_split:
    if line:
      software_reference, url = line.split(' ')
      software_product_list[software_reference] = url
  return software_product_list


class ProxyConfig(object):
  def __init__(self, logger):
    self.logger = logger
    self.multimaster = {}
    self.software_product_list = []

  def mergeConfig(self, args, configp):
    # Set arguments parameters (from CLI) as members of self
    for option, value in args.__dict__.items():
      setattr(self, option, value)

    for section in configp.sections():
      configuration_dict = dict(configp.items(section))
      if section in ("slapproxy", "slapos"):
        # Merge the arguments and configuration as member of self
        for key in configuration_dict:
          if not getattr(self, key, None):
            setattr(self, key, configuration_dict[key])
      elif section.startswith('multimaster/'):
        # Merge multimaster configuration if any
        # XXX: check for duplicate SR entries
        for key, value in six.iteritems(configuration_dict):
          if key == 'software_release_list':
            # Split multi-lines values
            configuration_dict[key] = [line.strip() for line in value.strip().split('\n')]
        self.multimaster[section.split('multimaster/')[1]] = configuration_dict

  def setConfig(self):
    if not self.database_uri:
      raise ValueError('database-uri is required.')
    # XXX: check for duplicate SR entries.
    self.software_product_list = _generateSoftwareProductListFromString(
        getattr(self, 'software_product_list', ''))


def setupFlaskConfiguration(conf):
  app.config['computer_id'] = conf.computer_id
  app.config['DATABASE_URI'] = conf.database_uri
  app.config['software_product_list'] = conf.software_product_list
  app.config['multimaster'] = conf.multimaster
  app.config['run_id'] = ''.join([random.choice(string.ascii_letters + string.digits) for n in range(32)])
  if getattr(conf, 'local_software_release_root', None) is not None:
    app.config['local_software_release_root'] = conf.local_software_release_root
  if getattr(conf, 'public_directory_path', None) is not None:
    app.config['PUBLIC_DIRECTORY_PATH'] = conf.public_directory_path

def connectDB():
  # if first connection, create an empty db at DATABASE_URI path
  conn = sqlite_connect(app.config['DATABASE_URI'])
  conn.close()

def do_proxy(conf):
  for handler in conf.logger.handlers:
    app.logger.addHandler(handler)

  from flask.logging import default_handler
  app.logger.removeHandler(default_handler)

  setupFlaskConfiguration(conf)
  connectDB()
  app.run(host=conf.host, port=int(conf.port), threaded=True)

